/* SPDX-License-Identifier: Zlib */

#ifndef PRINT_H
#define PRINT_H

#include "zathura.h"

/**
 * Opens a print dialog to print the current file
 *
 * @param zathura
 */
void print(zathura_t* zathura);

#endif // PRINT_H
